#include "AnVMSR.h"

#ifndef _I386_PROC_REG_H_
/* Inlined Write Register Functions */
static inline void wrmsr64(UInt32 msr, UInt64 val)
{
    UInt32 lo = ((UInt32)(val & 0xFFFFFFFF));
    UInt32 hi = ((UInt32)((val & 0xFFFFFFFF00000000) >> 32));

#ifdef DEBUG
    printf("WRMSR64: 0x%.8X - %.16llX\n", msr, (unsigned long long)val);
#endif /* DEBUG */

    wrmsr(msr, lo, hi);
}
#endif

static inline void wrpmc64(UInt32 pmc, UInt64 val)
{
    UInt32 lo = ((UInt32)(val & 0xFFFFFFFF));
    UInt32 hi = ((UInt32)((val & 0xFFFFFFFF00000000) >> 32));

#ifdef DEBUG
    printf("WRPMC64: 0x%.8X - %.16llX\n", pmc, (unsigned long long)val);
#endif /* DEBUG */

    wrpmc(pmc, lo, hi);
}

static inline void wrtsc64(UInt64 tsc)
{
    UInt32 lo = ((UInt32)(tsc & 0xFFFFFFFF));
    UInt32 hi = ((UInt32)((tsc & 0xFFFFFFFF00000000) >> 32));

#ifdef DEBUG
    UInt64 rv = (((UInt64)hi) << 32) | ((UInt64)lo);

    printf("WRTSC64: 0x%.16llX - %.16llX\n", (unsigned long long)tsc, (unsigned long long)rv);
#endif /* DEBUG */

    write_tsc(lo, hi);
}

#ifndef _I386_PROC_REG_H_
/* Inlined Read Register Functions */
static inline UInt64 rdtsc64(void)
{
    UInt32 lo = 0;
    UInt32 hi = 0;
    UInt64 rv = 0;

    rdtsc(lo, hi);

    rv = (((uint64_t)hi) << 32) | ((uint64_t)lo);

#ifdef DEBUG
    printf("RDTSC64: 0x%.16llX\n", ((unsigned long long)rv));
#endif /* DEBUG */

    return rv;
}
#endif /* _I386_PROC_REG_H_ */

#ifndef _I386_PROC_REG_H_
static inline UInt64 rdtscp64(UInt32 *aux)
{
    UInt32 lo = 0;
    UInt32 hi = 0;
    UInt64 rv = 0;

    if (aux == NULL)
    {
        printf("RDTSCP64 Error: Invalid Input Parameter!\n");
    } else {
        __asm__ volatile ("rdtscp; mov %%ecx, %1"
                          : "=a" (lo), "=d" (hi), "=m" (*aux)
                          :
                          : "ecx");

        rv = (((uint64_t)hi) << 32) | ((uint64_t)lo);

#ifdef DEBUG
        printf("RDTSCP64: 0x%.16llX - 0x%.16llX\n", ((unsigned long long)aux), ((unsigned long long)rv));
#endif /* DEBUG */
    }

    return rv;
}
#endif /* _I386_PROC_REG_H_ */

#ifndef _I386_PROC_REG_H_
static inline UInt64 rdmsr64(UInt32 msr)
{
    UInt32 lo = 0;
    UInt32 hi = 0;
    UInt64 rv = 0;

    rdmsr(msr, lo, hi);

    rv = (((uint64_t)hi) << 32) | ((uint64_t)lo);

#ifdef DEBUG
    printf("RDMSR64: 0x%.8X - 0x%.llX\n", msr, (unsigned long long)rv);
#endif /* DEBUG */

    return rv;
}
#endif /* _I386_PROC_REG_H_ */

#ifndef _I386_PROC_REG_H_
static inline UInt64 rdpmc64(UInt32 pmc)
{
    UInt32 lo = 0;
    UInt32 hi = 0;
    UInt64 rv = 0;

    rdpmc(pmc, lo, hi);

    rv = (((UInt64)hi) << 32) | ((UInt64)lo);

#ifdef DEBUG
    printf("RDPMC64: 0x%.8X - %.16llX\n", pmc, (unsigned long long)rv);
#endif /* DEBUG */

    return rv;
}
#endif /* _I386_PROC_REG_H_ */

#ifndef super
#define super IOService
#endif /* super */

OSDefineMetaClassAndStructors(AnVMSR, IOService)

bool AnVMSR::init(void)
{
    bool res = false;
#ifdef DEBUG
    UInt32 propCount = 0;
    UInt32 propCapacity = 0;
#endif /* DEBUG */

    mPropTable = this->getPropertyTable();

    res = ((bool)(mPropTable != NULL));

    if (res == false)
    {
        res = super::getPropertyTable();
    }

    if (res == true)
    {
#ifdef DEBUG
        propCount = mPropTable->getCount();
        propCapacity = mPropTable->getCapacity();

        printf("AnVMSR(%p) Found Kernel Extension IORegistry Properties: %u Entries Found, %u Entries Maximum.\n", this, propCount, propCapacity);
#endif /* DEBUG */
    } else {
#ifdef DEBUG
        printf("AnVMSR(%p) Found No Kernel Extension IORegistry Properties.\n", this);
#endif /* DEBUG */

        mPropTable = NULL;
    }

    if (mPropTable == NULL)
    {
        mPropTable = super::getPropertyTable();
    }

    if (mPropTable != NULL)
    {
        res = this->init(mPropTable);

        this->setProperties(mPropTable);
    } else {
          res = this->init(NULL);
    }

    mResService = this->getResourceService();

#ifdef  DEBUG
    printf("AnVMSR(%p) Initializing Kernel Extension With Default Property Table: %s.\n", this, ((res == true) ? "Succeeded" : "Failed"));
#endif /* DEBUG */

    return res;
}

bool AnVMSR::init(OSDictionary *dict)
{
    bool res = super::init(dict);

    mIORegID = this->getRegistryEntryID();
    mIORegEntry = this->getRegistryRoot();

#ifdef  DEBUG
    printf("AnVMSR(%p): Initializing Kernel Extension: %s.\n", this, ((res == true) ? "Succeeded" : "Failed"));
#endif /* DEBUG */

    return (res);
}

void AnVMSR::free(void)
{
#ifdef  DEBUG
    printf("AnVMSR(%p): Freeing Kernel Extension.\n", this);
#endif /* DEBUG */

    super::free();
}

bool AnVMSR::start(IOService *provider)
{
    bool res = super::start(provider);

    printf("AnVMSR(%p): Starting Kernel Extension - %s.\n", this, ((res == true) ? "Success" : "Failed"));

    if (res == true)
    {
#ifdef DONT_USE_ASYNCH
        registerService(0);
#else /* USE_ASYNCHRONOUS */
        registerService(kIOServiceAsynchronous);
#endif /* SYNCHRONOUS */

        mPrefPanelMemoryBufSize = BUFSIZE;
    }

    return res;
}

void AnVMSR::stop(IOService *provider)
{
    kprintf("AnVMSR(%p): Stopping Kernel Extension %p.\n", this, provider);

    super::stop(provider);
}

UInt64 AnVMSR::a_rdmsr(UInt32 msr)
{
    return rdmsr64(msr);
}

UInt64 AnVMSR::a_rdpmc(UInt32 pmc)
{
    return rdpmc64(pmc);
}

UInt64 AnVMSR::a_rdtsc(void)
{
    return rdtsc64();
}

void AnVMSR::a_wrmsr(UInt32 msr, UInt64 value)
{
    wrmsr64(msr, value);
}

void AnVMSR::a_wrpmc(UInt32 pmc, UInt64 value)
{
    wrpmc64(pmc, value);
}

void AnVMSR::a_wrtsc(UInt64 value)
{
    wrtsc64(value);
}

IOReturn AnVMSR::runAction(AnVMSRAction action, UInt32 * __unused outSize, void ** __unused outData, void * __unused extraArg)
{
#ifdef  DEBUG
    kprintf("Action: 0x%lx", ((unsigned long)action));
#else
    ((void)action);
#endif /* DEBUG */

	return kIOReturnSuccess;
}

IOReturn AnVMSR::newUserClient(task_t owningTask, void * __unused securityID, UInt32 type,
                               OSDictionary *properties, IOUserClient **handler)
{
    IOReturn retval = kIOReturnSuccess;
    AnVMSRUserClient *client = NULL;

    if (mClientCount > MAXUSERS)
    {
        kprintf("AnVMSR(%p) Client Already Created: Not Deleted.\n", this);

        retval = kIOReturnError;

        return retval;
    }

    client = (AnVMSRUserClient *)AnVMSRUserClient::withTask(owningTask);

    if (client == NULL)
    {
        retval = kIOReturnNoResources;

        kprintf("AnVMSR(%p) New User Client: Can't Create User Client.\n", this);
    } else {
        if (retval == kIOReturnSuccess)
        {
            // Start the client so it can accept requests.
            client->attach(this);

            if (client->start(this) == false)
            {
                retval = kIOReturnError;

                kprintf("AnVMSR(%p) New User Client: Can't Start User Client.\n", this);

                if (client != NULL)
                {
                    client->release();
                }
            }
        }
    }

    if (retval != kIOReturnSuccess)
    {
        kprintf("AnVMSR(%p) Start New User Client Error: %llu.\n", this, (unsigned long long)retval);

        client = NULL;
    } else {
        mClientPtr[mClientCount] = client;

        if (handler != NULL)
        {
            *handler = client;
        }

        if (client != NULL)
        {
            client->set_Q_Size(type);

            if (properties != NULL)
            {
                client->setPropertyTable(properties);
            }
        }

        mClientCount++;

#ifdef  DEBUG
        kprintf("AnVMSR(%p) Start New User Client: %p, User Client Count %llu.\n",
                this, mClientPtr[mClientCount], (unsigned long long)mClientCount);
#endif /* DEBUG */
    }

    return retval;
}

IOReturn AnVMSR::newUserClient(task_t owningTask, void * __unused securityID,
                               UInt32 type, IOUserClient **handler)
{
    IOReturn ioReturn = kIOReturnSuccess;
    AnVMSRUserClient *client = NULL;

    if (mClientCount > MAXUSERS)
    {
        kprintf("AnVMSR(%p): Client Already Created, Not Deleted.\n", this);

        ioReturn = kIOReturnError;

        return ioReturn;
    }

    client = (AnVMSRUserClient *)AnVMSRUserClient::withTask(owningTask);

    if (client == NULL)
    {
        ioReturn = kIOReturnNoResources;

        kprintf("AnVMSR(%p) New User Client: Can't Create User Client.\n", this);
    } else {
        if (ioReturn == kIOReturnSuccess)
        {
            // Start the client so it can accept requests.
            client->attach(this);

            if (client->start(this) == false)
            {
                ioReturn = kIOReturnError;

                kprintf("AnVMSR(%p) New User Client: Can't Start User Client.\n", this);

                if (client != NULL)
                {
                    client->release();
                }
            }
        }
    }

    if (ioReturn != kIOReturnSuccess)
    {
        kprintf("AnVMSR(%p) Start New User Client Error: %llu.\n", this, (unsigned long long)ioReturn);

        client = NULL;
    } else {
        mClientPtr[mClientCount] = client;

        if (handler != NULL)
        {
            *handler = client;
        }

        if (client != NULL)
        {
            client->set_Q_Size(type);
        }

        mClientCount++;

#ifdef  DEBUG
        kprintf("AnVMSR(%p) Start New User Client: %p, User Client Count %llu.\n", this,
               mClientPtr[mClientCount], (unsigned long long)mClientCount);
#endif /* DEBUG */
    }

    return ioReturn;
}

void AnVMSR::setErr(bool set)
{
#ifdef DEBUG
    kprintf("AnVMSR(%p) Set Error: %s.\n", this, ((set == true) ? "Yes" : "No"));
#endif /* DEBUG */

    mErrFlag = set;
}

bool AnVMSR::closeChild(AnVMSRUserClient *ptr)
{
    UInt64 i = 0;
    UInt64 idx = 0;
    bool retval = false;

    if (ptr == NULL)
    {
        kprintf("AnVMSR(%p) Close Child: Invalid Parameter Error.\n", this);

        return retval;
    }

    if (mClientCount == 0)
    {
        kprintf("AnVMSR(%p) Close Child: No Opened User Clients.\n", this);

        return retval;
    } else {
#ifdef DEBUG
        kprintf("AnVMSR(%p) Close Child: Find And Close User Client %p, %llu User Clients Checked.\n", this, ptr, (unsigned long long)mClientCount);
#endif /* DEBUG */

        for (i=0;i<mClientCount;i++)
        {
            if (mClientPtr[i] == ptr)
            {
                mClientCount--;

#ifdef DEBUG
                kprintf("AnVMSR(%p) Closed User Client: %p With Reference Nr %llu (%llu Opened User Clients).\n",
                       this, mClientPtr[i], (unsigned long long)i, (unsigned long long)mClientCount);
#endif /* DEBUG */

                mClientPtr[i] = NULL;
                idx = i;

                i = ((unsigned long long)(mClientCount + 1));

                retval = true;
            }
        }

        for (i=idx;i<mClientCount;i++)
        {
            mClientPtr[i] = mClientPtr[i+1];
        }

        mClientPtr[mClientCount+1] = NULL;
    }

    return retval;
}

/* Reserved Base Methods */
void AnVMSR::_RESERVEDAnVMSR0(void)
{
    kprintf("AnVMSRb(%p)::_RESERVEDAnVMSR%d Called.\n", this, 0);
}

void AnVMSR::_RESERVEDAnVMSR1(void)
{
    kprintf("AnVMSR(%p)::_RESERVEDAnVMSR%d Called.\n", this, 1);
}
                                                                    
void AnVMSR::_RESERVEDAnVMSR2(void)
{
    kprintf("AnVMSR(%p)::_RESERVEDAnVMSR%d Called.\n", this, 2);
}

// As these slots are used move them up inside the #if above
void AnVMSR::_RESERVEDAnVMSR3(void)
{
    kprintf("AnVMSR(%p)::_RESERVEDAnVMSR%d Called.\n", this, 3);
}
                                                                    
void AnVMSR::_RESERVEDAnVMSR4(void)
{
    kprintf("AnVMSR(%p)::_RESERVEDAnVMSR%d Called.\n", this, 4);
}

void AnVMSR::_RESERVEDAnVMSR5(void)
{
    kprintf("AnVMSR(%p)::_RESERVEDAnVMSR%d Called.\n", this, 5);
}

void AnVMSR::_RESERVEDAnVMSR6(void)
{
    kprintf("AnVMSR(%p)::_RESERVEDAnVMSR%d Called.\n", this, 6);
}

void AnVMSR::_RESERVEDAnVMSR7(void)
{
    kprintf("AnVMSR(%p)::_RESERVEDAnVMSR%d Called.\n", this, 7);
}

#undef  super
#define super IOUserClient

OSDefineMetaClassAndStructors(AnVMSRUserClient, IOUserClient);

const AnVMSRUserClient *AnVMSRUserClient::withTask(task_t owningTask)
{
    AnVMSRUserClient *client = new AnVMSRUserClient;

#ifdef DEBUG
    kprintf("AnVMSRUserClient New User Client: %p With Task %p.\n", client, owningTask);
#endif /* DEBUG */

    if (client != NULL)
	{
        if (client->init() == false)
		{
            client->release();

            client = NULL;
        } else {
            client->fTask = owningTask;
        }
    }

    return client;
}

bool AnVMSRUserClient::set_Q_Size(UInt64 capacity)
{
    if (capacity == 0)
    {
        return true;
    } else {
#ifdef  DEBUG
        kprintf("AnVMSRUserClient(%p): Resetting Data Queue Size And Setting Capacity To %llu Entries.\n", this, (unsigned long long)capacity);
#endif /* DEBUG */

        myLogQueue->release();

        /* Get memory for new queue of calcuated size */
#if USE_OLD_IODATAQUEUE == 1
        myLogQueue = new IODataQueue;
#else
        myLogQueue = new IOSharedDataQueue;
#endif

        if (myLogQueue == 0)
        {
            kprintf("AnVMSRUserClient(%p): Error Due To Memory Allocation Failure For Buffer.\n", this);

            Q_Err++;

            return false;
        }
    }

    if (myLogQueue == 0)
    {
        kprintf("AnVMSRUserClient(%p): Error Due To Memory Allocation Failure For Buffer.\n", this);

        Q_Err++;
    }

    if (myLogQueue->initWithCapacity((UInt32)capacity) == false)
    {
        kprintf("AnVMSRUserClient(%p): Error Couldn't Set Data Queue Size To %llu Entries.\n",
               this, ((unsigned long long)capacity));

        Q_Err++;

        return false;
    }

    return true;
}

void AnVMSRUserClient::messageHandler(UInt32 type, const char *format, ...)
{
    va_list args = { 0 };

#ifdef DEBUG
    kprintf("AnVMSRUserClient(%p) Message Handler: Type %u, Format %p.\n", this, type, format);
#else
    ((void)type);
#endif /* DEBUG */

	va_start(args, format);
	vprintf(format, args);
	va_end(args);
}

bool AnVMSRUserClient::initWithTask(task_t owningTask, void * securityToken, UInt32 type)
{
    OSDictionary *dict = this->getPropertyTable();

    return this->initWithTask(owningTask, securityToken, type, dict);
}

bool AnVMSRUserClient::MergeDictionaryIntoDictionary(OSDictionary *parentSourceDictionary, OSDictionary *parentTargetDictionary)
{
    OSCollectionIterator *srcIterator = NULL;
    OSSymbol *keyObject = NULL;
    OSDictionary *childSourceDictionary = NULL;
    OSDictionary *childTargetDictionary = NULL;
    OSObject *childTargetObject = NULL;
    bool result = false;
#ifdef DEBUG
    UInt64 mergedEntries = 0;
#endif /* DEBUG */

    if ((parentSourceDictionary == NULL) || (parentTargetDictionary == NULL))
    {
#ifdef DEBUG
        kprintf("AnVMSRUserClient(%p) Merge Dictionary Into Provider Parameter Error: Source Dictionary (%p), Target Dictionary (%p).\n",
               this, parentSourceDictionary, parentTargetDictionary);
#endif /* DEBUG */

        return result;
    }
#ifdef DEBUG
    else {
        kprintf("AnVMSRUserClient(%p) Merge Dictionary Into Dictionary: Source Dictionary %p, Target Dictionary %p.\n",
               this, parentSourceDictionary, parentTargetDictionary);
    }
#endif /* DEBUG */

    // Get our source dictionary
    srcIterator = OSCollectionIterator::withCollection(parentSourceDictionary);

    if (srcIterator != NULL)
    {
        while ((keyObject = OSDynamicCast(OSSymbol, srcIterator->getNextObject())) != NULL)
        {
            // Check to see if our destination already has the same entry.
            childTargetObject = parentTargetDictionary->getObject(keyObject);

            if (childTargetObject != NULL)
            {
                childTargetDictionary = OSDynamicCast(OSDictionary, childTargetObject);
            }

            // See if our source entry is also a dictionary
            childSourceDictionary = OSDynamicCast(OSDictionary, parentSourceDictionary->getObject(keyObject));

            if ((childTargetDictionary != NULL) && (childSourceDictionary != NULL))
            {
                // Our target dictionary already has the entry for this same object AND our
                // source is also a dictionary, so we need to recursively add it.
                // Need to merge our entry into the provider's dictionary.  However, we don't have a copy of our dictionary, just
                // a reference to it.  So, we need to make a copy of our target's dictionary
                if (childTargetDictionary == NULL)
                {
                    break;
                }
                    
                // Note that our targetDictionary *might* change
                // between the time we copied it and when we write it out again.  If so, we will obviously overwrite anychanges
                result = MergeDictionaryIntoDictionary(childSourceDictionary, childTargetDictionary);

                if (result == true)
                {
                    // Get the size of our provider's dictionary so to see if it's changed  (Yes, the size could remain the same but the contents
                    // could have changed, but this gives us a first approximation.  We're not doing anything with this result, although we could
                    // remerge
                    result = parentTargetDictionary->setObject(keyObject, childTargetDictionary);

                    if (result == false)
                    {
                        break;
                    }
#ifdef DEBUG
                    else {
                        ++mergedEntries;
                    }
#endif /* DEBUG */
                } else {
                    // If we got an error merging dictionaries, then just bail out without doing anything
                    break;
                }
            } else {
                // We have a property that we need to merge into our parent dictionary.
                result = parentTargetDictionary->setObject(keyObject, parentSourceDictionary->getObject(keyObject));

                if (result == false)
                {
                    break;
                }
#ifdef DEBUG
                else {
                    ++mergedEntries;
                }
#endif /* DEBUG */
            }
        }

        srcIterator->release();
    }

#ifdef DEBUG
    kprintf("AnVMSRUserClient(%p) Merge Dictionary Into Dictionary %s: Modified %llu Entries.\n",
           this, ((result == false) ? "Failed" : "Succeeded"), (unsigned long long)mergedEntries);
#endif /* DEBUG */

    return result;
}

bool AnVMSRUserClient::MergeDictionaryIntoProvider(IOService *provider, OSDictionary *dictionaryToMerge)
{
    const OSSymbol *dictionaryEntry = NULL;
    OSCollectionIterator *iter = NULL;
    bool result = false;
    OSDictionary *sourceDictionary = NULL;
    OSDictionary *providerDictionary = NULL;
    OSObject *providerProperty = NULL;
#ifdef DEBUG
    UInt64 mergedEntries = 0;
#endif /* DEBUG */

    if ((provider == NULL) || (dictionaryToMerge == NULL))
    {
#ifdef DEBUG
        kprintf("AnVMSRUserClient(%p) Merge Dictionary Into Provider: Invalid Parameter Error.\n", this);
#endif /* DEBUG */

        return result;
    }
#ifdef DEBUG
    else {
        kprintf("AnVMSRUserClient(%p) Merge Dictionary Into Provider: Provider %p, Dictionary %p.\n", this, provider, dictionaryToMerge);
    }
#endif /* DEBUG */
    
    // rdar://4041566 -- Trick the C++ run-time into keeping us loaded.
    if (haveCreatedRef == false)
    {
        haveCreatedRef = true;

        getMetaClass()->instanceConstructed();
    }
    
    // Get the dictionary whose entries we need to merge into our provider and get
    // an iterator to it.
    iter = OSCollectionIterator::withCollection((OSDictionary *)dictionaryToMerge);

    if (iter != NULL)
    {
        // Iterate through the dictionary until we run out of entries
        while ((dictionaryEntry = (const OSSymbol *)iter->getNextObject()) != NULL)
        {
            // Check to see if our destination already has the same entry.  If it does
            // we assume that it is a dictionary.  Perhaps we should check that
            providerProperty = provider->getProperty(dictionaryEntry);

            if (providerProperty != NULL)
            {
                providerDictionary = OSDynamicCast(OSDictionary, providerProperty);
            }

            // See if our source entry is also a dictionary
            //
            sourceDictionary = OSDynamicCast(OSDictionary, dictionaryToMerge->getObject(dictionaryEntry));

            if ((providerDictionary != NULL) && (sourceDictionary != NULL))
            {
                // Note that our providerDictionary *might* change
                // between the time we copied it and when we write it out again.  If so, we will obviously overwrite anychanges
                result = MergeDictionaryIntoDictionary(sourceDictionary, providerDictionary);

                if (result == true)
                {
                    // Get the size of our provider's dictionary so to see if it's changed  (Yes, the size could remain the same but the contents
                    // could have changed, but this gives us a first approximation.  We're not doing anything with this result, although we could
                    // remerge
                    result = provider->setProperty(dictionaryEntry, providerDictionary);

                    if (result == false)
                    {
                        break;
                    }
#ifdef DEBUG
                    else {
                        ++mergedEntries;
                    }
#endif /* DEBUG */
                } else {
                    // If we got an error merging dictionaries, then just bail out without doing anything
                    break;
                }
           } else {
                result = provider->setProperty(dictionaryEntry,
                                               dictionaryToMerge->getObject(dictionaryEntry));

                if (result == false)
                {
                    break;
                }
#ifdef DEBUG
                else {
                    ++mergedEntries;
                }
#endif /* DEBUG */
            }
        }

        iter->release();
    }

#ifdef DEBUG
    kprintf("AnVMSRUserClient(%p) Merge Dictionary Into Provider: %s (Added %llu Entries).\n",
            this, ((result == false) ? "Failed" : "Succeeded"), ((unsigned long long)mergedEntries));
#endif /* DEBUG */

    return result;
}

bool AnVMSRUserClient::initWithTask(task_t owningTask, void *securityID,
                                    UInt32 type, OSDictionary *properties)
{
    bool retval = super::initWithTask(owningTask, securityID, type, properties);

    if (properties != NULL)
    {
        this->MergeDictionaryIntoProvider(this->getProvider(), properties);
        this->setPropertyTable(properties);

#ifdef DEBUG
        kprintf("AnVMSRUserClient(%p) Merged Properties Dictionary: %p.\n", this, properties);
#endif /* DEBUG */
    }

#ifdef DEBUG
    kprintf("AnVMSRUserClient(%p) Initialize With Task: Owning Task %p, Security ID %p, Type %u, Properties %p (%s).\n",
            this, owningTask, securityID, type, properties, ((retval == false) ? "Failed" : "Succeeded"));
#endif /* DEBUG */

    return retval;
}

bool AnVMSRUserClient::start(IOService *provider)
{
    bool retval = super::start(provider);

#ifdef DEBUG
    kprintf("AnVMSRUserClient(%p): Start User Client %p (%s).\n",
            this, provider, ((retval == false) ? "Failed" : "Succeeded"));
#endif /* DEBUG */

	if (retval != false)
    {
        mDevice = OSDynamicCast(AnVMSR, provider);

        mDevice->retain();
    }

	return retval;
}

bool AnVMSRUserClient::willTerminate(IOService *provider, IOOptionBits options)
{
    bool retval = super::willTerminate(provider, options);

#ifdef DEBUG
    kprintf("AnVMSRUserClient(%p) Will Terminate: Provider %p, Options %llu (%s).\n",
            this, provider, (unsigned long long)options, ((retval == false) ? "Failed" : "Succeeded"));
#endif /* DEBUG */

	return retval;
}

bool AnVMSRUserClient::didTerminate(IOService *provider, IOOptionBits options, bool *defer)
{
    /* If defer is true, stop will not be called on the user client. */
    bool retval = super::didTerminate(provider, options, defer);

#ifdef DEBUG
    kprintf("AnVMSRUserClient(%p) Did Terminate: Provider -> %p, Options -> %llu, Is Deferred -> %s (%s).\n",
            this, provider, (unsigned long long)options,
            ((deferred == false) ? "No" : "Yes"), ((retval == false) ? "Unsuccessful" : "Successful"));
#endif /* DEBUG */

    return retval;
}

bool AnVMSRUserClient::terminate(IOOptionBits options)
{
    bool retval = super::terminate(options);

#ifdef DEBUG
    kprintf("AnVMSRUserClient(%p) Terminate With Options: %llu (%s).\n",
           this, (unsigned long long)options, ((retval == false) ? "Unsucessful" : "Succesful"));
#endif /* DEBUG */

    return retval;
}

/* clientClose is called when the user process calls IOServiceClose */
IOReturn AnVMSRUserClient::clientClose(void)
{
    IOReturn retval = kIOReturnSuccess;
    bool childClosed = false;

#ifdef DEBUG
    kprintf("AnVMSRUserClient(%p) Close User Client.\n", this);
#endif /* DEBUG */

    if (mDevice != NULL)
	{
        childClosed = mDevice->closeChild(this);

        kprintf("AnVMSRUserClient(%p) Closed Client Child: %s.\n", this,
               ((childClosed == true) ? "Yes" : "No"));
    }

	if (isInactive() == false)
    {
		retval = ((terminate(kIOServiceTerminate) == true) ? kIOReturnSuccess : kIOReturnError);

#ifdef DEBUG
        kprintf("AnVMSRUserClient(%p) Terminated User Client: %s.\n",
                this, ((retval == kIOReturnSuccess) ? "Yes" : "No"));
#endif /* DEBUG */
    }

	return retval;
}

// clientDied is called when the user process terminates unexpectedly, the default
// implementation simply calls clientClose
IOReturn AnVMSRUserClient::clientDied(void)
{
    return this->clientClose();
}

void AnVMSRUserClient::free(void)
{
#ifdef DEBUG
    kprintf("AnVMSRUserClient(%p) Free User Client.\n", this);
#endif /* DEBUG */

	mDevice->release();

	super::free();
}

/*
 * Stop will be called during the termination
 * process, and should free all resources
 * associated with this client.
 */
void AnVMSRUserClient::stop(IOService *provider)
{
#ifdef DEBUG
    kprintf("AnVMSRUserClient(%p) Stop: Provider %p.\n", this, provider);
#endif /* DEBUG */

	super::stop(provider);
}

IOExternalAsyncMethod *AnVMSRUserClient::getAsyncTargetAndMethodForIndex(IOService **target, UInt32 index)
{
    static const IOExternalAsyncMethod methodDescs[AnVMSRNumMethods+1] = {
        { NULL, ((IOAsyncMethod)&AnVMSRUserClient::actionMethodAsyncRDMSR),
            kIOUCStructIStructO, kIOUCVariableStructureSize, kIOUCVariableStructureSize },
        { NULL, ((IOAsyncMethod)&AnVMSRUserClient::actionMethodAsyncWRMSR),
            kIOUCStructIStructO, kIOUCVariableStructureSize, kIOUCVariableStructureSize },
        { NULL, ((IOAsyncMethod)&AnVMSRUserClient::actionMethodAsyncRDPMC),
            kIOUCStructIStructO, kIOUCVariableStructureSize, kIOUCVariableStructureSize },
        { NULL, ((IOAsyncMethod)&AnVMSRUserClient::actionMethodAsyncWRPMC),
            kIOUCStructIStructO, kIOUCVariableStructureSize, kIOUCVariableStructureSize },
        { NULL, ((IOAsyncMethod)&AnVMSRUserClient::actionMethodAsyncRDTSC),
            kIOUCStructIStructO, kIOUCVariableStructureSize, kIOUCVariableStructureSize },
        { NULL, ((IOAsyncMethod)&AnVMSRUserClient::actionMethodAsyncWRTSC),
            kIOUCStructIStructO, kIOUCVariableStructureSize, kIOUCVariableStructureSize },
        { NULL, ((IOAsyncMethod)0), 0, 0, 0 }
    };

    if (target != NULL)
    {
        *target = this;
    }

    if (index < AnVMSRNumMethods)
    {
        return ((IOExternalAsyncMethod *)(methodDescs + index));
    }

    return NULL;
}

/*
 * getTargetAndMethodForIndex looks
 * up the external methods - supply
 * a description of the parameters
 * available to be called.
 */
IOExternalMethod *AnVMSRUserClient::getTargetAndMethodForIndex(IOService **target, UInt32 index)
{
	static const IOExternalMethod methodDescs[AnVMSRNumMethods+1] = {
        { NULL, ((IOMethod)&AnVMSRUserClient::actionMethodRDMSR), kIOUCStructIStructO,
            kIOUCVariableStructureSize, kIOUCVariableStructureSize },
        { NULL, ((IOMethod)&AnVMSRUserClient::actionMethodWRMSR), kIOUCStructIStructO,
            kIOUCVariableStructureSize, kIOUCVariableStructureSize },
        { NULL, ((IOMethod)&AnVMSRUserClient::actionMethodRDPMC), kIOUCStructIStructO,
            kIOUCVariableStructureSize, kIOUCVariableStructureSize },
        { NULL, ((IOMethod)&AnVMSRUserClient::actionMethodWRPMC), kIOUCStructIStructO,
            kIOUCVariableStructureSize, kIOUCVariableStructureSize },
        { NULL, ((IOMethod)&AnVMSRUserClient::actionMethodRDTSC), kIOUCStructIStructO,
            kIOUCVariableStructureSize, kIOUCVariableStructureSize },
        { NULL, ((IOMethod)&AnVMSRUserClient::actionMethodWRTSC), kIOUCStructIStructO,
            kIOUCVariableStructureSize, kIOUCVariableStructureSize },
        { NULL, ((IOMethod)0), 0, 0, 0 }
	};

    if (target != NULL)
    {
        *target = this;
    }

	if (index < AnVMSRNumMethods)
    {
		return ((IOExternalMethod *)(methodDescs + index));
    }

    return NULL;
}

IOReturn AnVMSRUserClient::actionMethodAsyncRDMSR(OSAsyncReference __unused asyncRef,
                                                  UInt32 *dataIn, UInt32 *dataOut,
                                                  IOByteCount inputSize, IOByteCount *outputSize)
{
    return this->actionMethodRDMSR(dataIn, dataOut, inputSize, outputSize);
}

IOReturn AnVMSRUserClient::actionMethodAsyncRDPMC(OSAsyncReference __unused asyncRef,
                                                  UInt32 *dataIn, UInt32 *dataOut,
                                                  IOByteCount inputSize, IOByteCount *outputSize)
{
    return this->actionMethodRDPMC(dataIn, dataOut, inputSize, outputSize);
}

IOReturn AnVMSRUserClient::actionMethodAsyncRDTSC(OSAsyncReference __unused asyncRef,
                                                  UInt32 *dataIn, UInt32 *dataOut,
                                                  IOByteCount inputSize, IOByteCount *outputSize)
{
    return this->actionMethodRDTSC(dataIn, dataOut, inputSize, outputSize);
}

IOReturn AnVMSRUserClient::actionMethodAsyncWRMSR(OSAsyncReference __unused asyncRef,
                                                  UInt32 *dataIn, UInt32 *dataOut,
                                                  IOByteCount inputSize, IOByteCount *outputSize)
{
    return this->actionMethodWRMSR(dataIn, dataOut, inputSize, outputSize);
}

IOReturn AnVMSRUserClient::actionMethodAsyncWRPMC(OSAsyncReference __unused asyncRef,
                                                  UInt32 *dataIn, UInt32 *dataOut,
                                                  IOByteCount inputSize, IOByteCount *outputSize)
{
    return this->actionMethodWRPMC(dataIn, dataOut, inputSize, outputSize);
}

IOReturn AnVMSRUserClient::actionMethodAsyncWRTSC(OSAsyncReference __unused asyncRef,
                                                  UInt32 *dataIn, UInt32 *dataOut,
                                                  IOByteCount inputSize, IOByteCount *outputSize)
{
    return this->actionMethodWRTSC(dataIn, dataOut, inputSize, outputSize);
}

IOReturn AnVMSRUserClient::actionMethodRDMSR(UInt32 *dataIn, UInt32 *dataOut,
                                             IOByteCount __unused  inputSize, IOByteCount * __unused outputSize)
{
    msr_inout *msrdata = (msr_inout *)dataIn;
    msr_inout *msroutdata = (msr_inout *)dataOut;

    if ((msrdata == NULL) || (msroutdata == NULL))
    {
#ifdef DEBUG
        kprintf("AnVMSRUserClient(%p) Read MSR Error: Invalid Input (%p) Or Output (%p) Data Pointer.\n", this, msrdata, msroutdata);
#endif /* DEBUG */

        return kIOReturnBadArgument;
    }

#ifdef  DEBUG
    kprintf("AnVMSRUserClient(%p) Read MSR: Input Data %u Of %u Bytes, Output Data %u Of %u Bytes.\n",
            this, ((dataIn != NULL) ? *dataIn : 0), inputSize,
            ((dataOut != NULL) ? *dataOut : 0), ((outputSize != NULL) ? *outputSize : 0));
#endif /* DEBUG */

	msrdata->param = mDevice->a_rdmsr(msrdata->msr);

#ifdef  DEBUG
    kprintf("AnVMSRUserClient Read MSR Register 0x%X: 0x%llX.\n",
            msrdata->msr, (unsigned long long)msrdata->param);
#endif /* DEBUG */

    msroutdata->param = msrdata->param;

    return kIOReturnSuccess;
}

IOReturn AnVMSRUserClient::actionMethodWRMSR(UInt32 *dataIn, UInt32 * __unused dataOut,
                                             IOByteCount __unused inputSize,
                                             IOByteCount *__unused outputSize)
{
    msr_inout *msrdata = ((msr_inout *)dataIn);

#ifdef  DEBUG
    kprintf("AnVMSRUserClient(%p): Write MSR called.\n", this);
#endif /* DEBUG */

    if (msrdata == NULL)
    {
#ifdef DEBUG
        kprintf("AnVMSRUserClient(%p) Write MSR Error: Invalid Input (%p) Data Pointer.\n", this, msrdata);
#endif /* DEBUG */

        return kIOReturnUnsupported;
    }

#ifdef DEBUG
	kprintf("AnVMSRUserClient(%p)::actionMethod(%llu, %llu).\n",
            this, (unsigned long long)inputSize,
            ((outputSize != NULL) ?  (unsigned long long)*outputSize : 0));
#endif /* DEBUG */

	mDevice->a_wrmsr(msrdata->msr, msrdata->param);

#ifdef  DEBUG
    kprintf("AnVMSRUserClient(%p): WRMSR 0x%llX to %X.\n",
            this, ((unsigned long long)msrdata->param), msrdata->msr);
#endif

    return kIOReturnSuccess;
}

IOReturn AnVMSRUserClient::actionMethodRDPMC(UInt32 *dataIn, UInt32 *dataOut,
                                             IOByteCount __unused inputSize, IOByteCount *__unused outputSize)
{
    msr_inout *msrdata = (msr_inout *)dataIn;
    msr_inout *msroutdata = (msr_inout *)dataOut;

    if ((msrdata == NULL) || (msroutdata == NULL))
    {
#ifdef DEBUG
        kprintf("AnVMSRUserClient(%p) Read PMC Error: Invalid Input (%p) Or Output (%p) Data Pointer.\n",
                this, msrdata, msroutdata);
#endif /* DEBUG */

        return kIOReturnBadArgument;
    }

#ifdef  DEBUG
    kprintf("AnVMSRUserClient(%p) Read PMC: Input Data %u Of %llu Bytes, Output Data %u Of %llu Bytes.\n",
            this, ((dataIn != NULL) ? *dataIn : 0),
            (unsigned long long)inputSize, ((dataOut != NULL) ? *dataOut : 0),
            ((outputSize != NULL) ? (unsigned long long)*outputSize : 0));
#endif /* DEBUG */

    msrdata->param = mDevice->a_rdpmc(msrdata->msr);

#ifdef  DEBUG
    kprintf("AnVMSRUserClient(%p) Read PMC Register 0x%X: 0x%llX.\n",
            this, msrdata->msr, (unsigned long long)msrdata->param);
#endif /* DEBUG */

    msroutdata->param = msrdata->param;

    return kIOReturnSuccess;
}

IOReturn AnVMSRUserClient::actionMethodRDTSC(UInt32 *dataIn, UInt32 *dataOut,
                                             IOByteCount __unused inputSize, IOByteCount *__unused outputSize)
{
    msr_inout *msrdata = (msr_inout *)dataIn;
    msr_inout *msroutdata = (msr_inout *)dataOut;

    if ((msrdata == NULL) || (msroutdata == NULL))
    {
#ifdef DEBUG
        kprintf("AnVMSRUserClient(%p) Read TSC Error: Invalid Input (%p) Or Output (%p) Data Pointer.\n",
                this, msrdata, msroutdata);
#endif /* DEBUG */

        return kIOReturnBadArgument;
    }

#ifdef  DEBUG
    kprintf("AnVMSRUserClient(%p) Read TSC: Input Data %u Of %llu Bytes, Output Data %u Of %llu Bytes.\n",
            this, ((dataIn != NULL) ? *dataIn : 0), (unsigned long long)inputSize,
            ((dataOut != NULL) ? *dataOut : 0),
            ((outputSize != NULL) ? (unsigned long long)*outputSize : 0));
#endif /* DEBUG */

    msrdata->param = mDevice->a_rdtsc();

#ifdef  DEBUG
    kprintf("AnVMSRUserClient(%p) Read TSC Register 0x%X: 0x%llX.\n",
           this, msrdata->msr, (unsigned long long)msrdata->param);
#endif /* DEBUG */

    msroutdata->param = msrdata->param;

    return kIOReturnSuccess;
}

IOReturn  AnVMSRUserClient::actionMethodWRPMC(UInt32 *dataIn, UInt32 * __unused dataOut,
                                              IOByteCount __unused inputSize,
                                              IOByteCount * __unused outputSize)
{
    msr_inout *msrdata = ((msr_inout *)dataIn);

#ifdef  DEBUG
    kprintf("AnVMSRUserClient(%p): Write PMC called.\n", this);
#endif /* DEBUG */

    if (msrdata == NULL)
    {
#ifdef DEBUG
        kprintf("AnVMSRUserClient(%p) Write PMC Error: Invalid Input (%p) Data Pointer.\n", this, msrdata);
#endif /* DEBUG */

        return kIOReturnUnsupported;
    }

#ifdef DEBUG
    kprintf("AnVMSRUserClient(%p)::actionMethod(%llu, %llu).\n",
            this, (unsigned long long)inputSize,
            ((outputSize != NULL) ?  (unsigned long long)*outputSize : 0));
#endif /* DEBUG */

    mDevice->a_wrpmc(msrdata->msr, msrdata->param);

#ifdef  DEBUG
    kprintf("AnVMSRUserClient(%p): WRPMC 0x%llX to %X.\n",
            this, ((unsigned long long)msrdata->param), msrdata->msr);
#endif /* DEBUG */

    return kIOReturnSuccess;
}

IOReturn AnVMSRUserClient::actionMethodWRTSC(UInt32 *dataIn, UInt32 * __unused dataOut,
                                             IOByteCount __unused inputSize,
                                             IOByteCount * __unused outputSize)
{
    msr_inout *msrdata = ((msr_inout *)dataIn);

#ifdef  DEBUG
    kprintf("AnVMSRUserClient(%p): WRTSC called.\n", this);
#endif /* DEBUG */

    if (msrdata == NULL)
    {
        return kIOReturnUnsupported;
    }

#ifdef DEBUG
    kprintf("AnVMSRUserClient(%p)::actionMethod(%llu, %llu).\n",
            this, (unsigned long long)inputSize, (unsigned long long)*outputSize);
#endif /* DEBUG */

    mDevice->a_wrtsc(msrdata->param);
        
#ifdef  DEBUG
    kprintf("AnVMSRUserClient(%p): WRTSC 0x%.16llX.\n", this, ((unsigned long long)msrdata->param));
#endif /* DEBUG */

    return kIOReturnSuccess;
}

IOReturn AnVMSRUserClient::clientMemoryForType(UInt32 __unused type,
                                               IOOptionBits *options,
                                               IOMemoryDescriptor **memory)
{
	//IOBufferMemoryDescriptor *memDesc = NULL;
	//void *msgBuffer = NULL;

    if (options)
    {
        *options = 0;
    }

    if (memory)
    {
        *memory = NULL;
    }

#ifdef  DEBUG
    kprintf("AnVMSRUserClient: Client Memory For Type Called.\n");
#endif /* DEBUG */

    memDesc = IOBufferMemoryDescriptor::withOptions(mDevice->getMemorySize(),
                                                    kIODirectionOutIn,
                                                    false);

    if (memDesc)
    {
        memDesc->appendBytes(mDevice->getMemoryBuf(), mDevice->getMemorySize());

        /*
         * msgBuffer = (char *)memDesc->getBytesNoCopy();
         * memcpy(msgBuffer, mDevice->getMemoryBuf(), mDevice->getMemorySize());
         */
        if (memory)
        {
            *memory = memDesc; // automatically released after memory is mapped into task
        }
    } else {
#ifdef  DEBUG
        kprintf("AnVMSRUserClient: Invalid Buffer Memory Descriptor.\n");
#endif /* DEBUG */

        return kIOReturnUnsupported;
    }

    return kIOReturnSuccess;
}

/* Reserved Base Methods */
void AnVMSRUserClient::_RESERVEDAnVMSRUserClient0(void)
{
    kprintf("AnVMSRUserClient(%p)::_RESERVEDAnVMSRUserClient%d Called.\n", this, 0);
}

void AnVMSRUserClient::_RESERVEDAnVMSRUserClient1(void)
{
    kprintf("AnVMSRUserClient(%p)::_RESERVEDAnVMSRUserClient%d Called.\n", this, 1);
}
                                  
void AnVMSRUserClient::_RESERVEDAnVMSRUserClient2(void)
{
    kprintf("AnVMSRUserClient(%p)::_RESERVEDAnVMSRUserClient%d Called.\n", this, 2);
}

// As these slots are used move them up inside the #if above
void AnVMSRUserClient::_RESERVEDAnVMSRUserClient3(void)
{
    kprintf("AnVMSRUserClient(%p)::_RESERVEDAnVMSRUserClient%d Called.\n", this, 3);
}
                                  
void AnVMSRUserClient::_RESERVEDAnVMSRUserClient4(void)
{
    kprintf("AnVMSRUserClient(%p)::_RESERVEDAnVMSRUserClient%d Called.\n", this, 4);
}

void AnVMSRUserClient::_RESERVEDAnVMSRUserClient5(void)
{
    kprintf("AnVMSRUserClient(%p)::_RESERVEDAnVMSRUserClient%d Called.\n", this, 5);
}

void AnVMSRUserClient::_RESERVEDAnVMSRUserClient6(void)
{
    kprintf("AnVMSRUserClient(%p)::_RESERVEDAnVMSRUserClient%d Called.\n", this, 6);
}

void AnVMSRUserClient::_RESERVEDAnVMSRUserClient7(void)
{
    kprintf("AnVMSRUserClient(%p)::_RESERVEDAnVMSRUserClient%d Called.\n", this, 7);
}
