/*
 *  main.m
 *  anvmsr
 *
 *  Created by Andy Vandijck on 12/09/13.
 *  Copyright (c) 2013-2019 Andy Vandijck. All rights reserved.
 */

#import <Foundation/Foundation.h>

#ifndef kAnVMSRClassName
#define kAnVMSRClassName "AnVMSR"
#endif /* kAnVMSRClassName */

io_service_t getService(void);
void usage(const char *name);
long long hex2int(const char *s);

typedef enum _anvmsr_methods
{
    AnVMSRActionMethodRDMSR = 0,
    AnVMSRActionMethodWRMSR = 1,
    AnVMSRActionMethodRDPMC = 2,
    AnVMSRActionMethodWRPMC = 3,
    AnVMSRActionMethodRDTSC = 4,
    AnVMSRActionMethodWRTSC = 5,
    AnVMSRNumMethods /* 6 Methods */
} anvmsr_methods_t;

typedef struct _msr_inout
{
	UInt32 action;
    UInt32 msr;

    UInt64 param;
} msr_inout;

io_service_t getService(void)
{
	io_service_t service = (io_service_t)0;
	mach_port_t masterPort = (mach_port_t)0;
	io_iterator_t iter = (io_iterator_t)0;
	kern_return_t ret = (kern_return_t)0;
    io_string_t path = { 0 };

	ret = IOMasterPort(MACH_PORT_NULL, &masterPort);

	if (ret != KERN_SUCCESS)
    {
		fprintf(stderr, "Can't get masterport.\n");

		return service;
	}
	
	ret = IOServiceGetMatchingServices(masterPort, IOServiceMatching(kAnVMSRClassName), &iter);

	if (ret != KERN_SUCCESS)
    {
		fprintf(stderr, "AnVMSR driver is not running.\n");

		return service;
	}

	service = IOIteratorNext(iter);
	IOObjectRelease(iter);

	ret = IORegistryEntryGetPath(service, kIOServicePlane, path);

	if (ret != KERN_SUCCESS)
    {
		fprintf(stderr, "Can't get the registry entry path.\n");
	}

    return service;
}

void usage(const char *name)
{
    printf("AnVMSR Version %s, Copyright (C) 2013-2019 AnV Software.\n\n", ANVMSRVER);
    printf("Usage:\n");
    printf("Read  MSR: %s readmsr <HEX_MSR>\n", name);
    printf("Write MSR: %s writemsr <HEX_MSR> <HEX_VALUE>\n", name);
    printf("Read  PMC: %s readpmc <HEX_PMC>\n", name);
    printf("Write PMC: %s writepmc <HEX_PMC> <HEX_VALUE>\n", name);
    printf("Read  TSC: %s readtsc\n", name);
    printf("Write TSC: %s writetsc <HEX_VALUE>\n\n", name);
    printf(" <HEX_MSR>  is hex without 0x prefix\n");
    printf(" <HEX_PMC>  is hex without 0x prefix\n");
    printf("<HEX_VALUE> is hex without 0x prefix\n");
}

long long hex2int(const char *s)
{
    const char charset[16] = "0123456789abcdef";
    const char altcharset[16] = "0123456789ABCDEF";
    long long i = (long long)strlen(s);
    long long len = i;
    long long num = 0;
    long long j = 0;

    while (i >= 0)
    {
        for (j = 0; j < 16; j++)
        {
            if ((charset[j] == s[i]) || (altcharset[j] == s[i]))
            {
                num += (j * (long long)pow(16, ((len - i) - 1)));

                break;
            }
        }

        i--;
    }

    return (num);
}

int main(int argc, const char **argv)
{
    char *parameter = NULL;
    char *msr = NULL;
    char *regvalue = NULL;
    io_service_t service = getService();
    io_connect_t connect = (io_connect_t)0;
    msr_inout msrin = { 0, 0, 0 };
    msr_inout msrout = { 0, 0, 0 };
    size_t outsize = 0;
    kern_return_t ret = 0;
    int retval = 0;

	if (service == ((io_service_t)0))
    {
        return -1;
    }

    outsize = sizeof(msrout);

    if (argc >= 2)
    {
        parameter = (char *)argv[1];
    } else {
        usage(argv[0]);

        retval = 1;

        return retval;
    }

    if (argc >= 3)
    {
        msr = (char *)argv[2];
    } else {
        msr = NULL;
    }

    ret = IOServiceOpen(service, mach_task_self(), 0, &connect);

    if (ret != KERN_SUCCESS)
    {
        fprintf(stderr, "Couldn't open IOService.\n");

        retval = -1;

        return retval;
    }

    if (!strncmp(parameter, "readmsr", 7))
    {
        if (argc < 3)
        {
            usage(argv[0]);

            retval = 1;

            return retval;
        }

        msrin.msr = (UInt32)hex2int(msr);
        msrin.action = AnVMSRActionMethodRDMSR;
        msrin.param = 0;

#if MAC_OS_X_VERSION_MIN_REQUIRED <= 1040
        ret = IOConnectMethodStructureIStructureO(connect,
                                                  AnVMSRActionMethodRDMSR,
                                                  sizeof(msrin), /* structureInputSize */
                                                  &outsize, /* structureOutputSize */
                                                  &msrin, /* inputStructure */
                                                  &msrout); /* ouputStructure */
#else /* MAC_OS_X_VERSION_MIN_REQUIRED > 1040 */
        ret = IOConnectCallStructMethod(connect,
                                        AnVMSRActionMethodRDMSR,
                                        &msrin,
                                        sizeof(msrin),
                                        &msrout,
                                        &outsize);
#endif /* MAC_OS_X_VERSION_MIN_REQUIRED */

        if (ret != KERN_SUCCESS)
        {
            fprintf(stderr, "Can't connect to structure method to send commands.\n");

            retval = -1;
        } else {
            printf("RDMSR 0x%x returns value 0x%llx.\n", (unsigned int)msrin.msr, (unsigned long long)msrout.param);
        }
    } else if (!strncmp(parameter, "readpmc", 7)) {
        if (argc < 3)
        {
            usage(argv[0]);

            retval = 1;

            return retval;
        }

        msrin.msr = (UInt32)hex2int(msr);
        msrin.action = AnVMSRActionMethodRDPMC;
        msrin.param = 0;

#if MAC_OS_X_VERSION_MIN_REQUIRED <= 1040
        ret = IOConnectMethodStructureIStructureO(connect,
                                                  AnVMSRActionMethodRDPMC,
                                                  sizeof(msrin), /* structureInputSize */
                                                  &outsize, /* structureOutputSize */
                                                  &msrin, /* inputStructure */
                                                  &msrout); /* ouputStructure */
#else /* MAC_OS_X_VERSION_MIN_REQUIRED > 1040 */
        ret = IOConnectCallStructMethod(connect,
                                        AnVMSRActionMethodRDPMC,
                                        &msrin,
                                        sizeof(msrin),
                                        &msrout,
                                        &outsize);
#endif /* MAC_OS_X_VERSION_MIN_REQUIRED */

        if (ret != KERN_SUCCESS)
        {
            fprintf(stderr, "Can't connect to structure method to send commands.\n");

            retval = -1;
        } else {
            printf("RDPMC 0x%x returns value 0x%llx.\n", (unsigned int)msrin.msr, (unsigned long long)msrout.param);
        }
    } else if (!strncmp(parameter, "readtsc", 7)) {
        if (argc < 2)
        {
            usage(argv[0]);

            retval = 1;

            return retval;
        }

        msrin.msr = 0;
        msrin.action = AnVMSRActionMethodRDTSC;
        msrin.param = 0;

#if MAC_OS_X_VERSION_MIN_REQUIRED <= 1040
        ret = IOConnectMethodStructureIStructureO(connect,
                                                  AnVMSRActionMethodRDTSC,
                                                  sizeof(msrin), /* structureInputSize */
                                                  &outsize, /* structureOutputSize */
                                                  &msrin, /* inputStructure */
                                                  &msrout); /* ouputStructure */
#else /* MAC_OS_X_VERSION_MIN_REQUIRED > 1040 */
        ret = IOConnectCallStructMethod(connect,
                                        AnVMSRActionMethodRDTSC,
                                        &msrin,
                                        sizeof(msrin),
                                        &msrout,
                                        &outsize);
#endif /* MAC_OS_X_VERSION_MIN_REQUIRED */

        if (ret != KERN_SUCCESS)
        {
            fprintf(stderr, "Can't connect to structure method to send commands.\n");

            retval = -1;
        } else {
            printf("RDTSC returns value 0x%llx.\n", (unsigned long long)msrout.param);
        }
    } else if (!strncmp(parameter, "writemsr", 8)) {
        if (argc < 4)
        {
            usage(argv[0]);

            retval = 1;

            return retval;
        }

        regvalue = (char *)argv[3];

        msrin.msr = (UInt32)hex2int(msr);
        msrin.action = AnVMSRActionMethodWRMSR;
        msrin.param = (UInt64)hex2int(regvalue);

#if MAC_OS_X_VERSION_MIN_REQUIRED <= 1040
        ret = IOConnectMethodStructureIStructureO(connect,
                                                  AnVMSRActionMethodWRMSR,
                                                  sizeof(msrin),			/* structureInputSize */
                                                  &outsize,    /* structureOutputSize */
                                                  &msrin,        /* inputStructure */
                                                  &msrout);       /* ouputStructure */
#else /* MAC_OS_X_VERSION_MIN_REQUIRED > 1040 */
        ret = IOConnectCallStructMethod(connect,
                                        AnVMSRActionMethodWRMSR,
                                        &msrin,
                                        sizeof(msrin),
                                        &msrout,
                                        &outsize);
#endif /* MAC_OS_X_VERSION_MIN_REQUIRED */

        if (ret != KERN_SUCCESS)
        {
            fprintf(stderr, "Can't connect to structure method to send commands.\n");

            retval = -1;
        } else {
            printf("WRMSR 0x%x with value 0x%llx.\n", (unsigned int)msrin.msr, (unsigned long long)msrin.param);
        }
    } else if (!strncmp(parameter, "writepmc", 8)) {
        if (argc < 4)
        {
            usage(argv[0]);

            retval = 1;

            return retval;
        }

        regvalue = (char *)argv[3];

        msrin.msr = (UInt32)hex2int(msr);
        msrin.action = AnVMSRActionMethodWRPMC;
        msrin.param = (UInt64)hex2int(regvalue);

#if MAC_OS_X_VERSION_MIN_REQUIRED <= 1040
        ret = IOConnectMethodStructureIStructureO(connect,
                                                  AnVMSRActionMethodWRPMC,
                                                  sizeof(msrin),            /* structureInputSize */
                                                  &outsize,    /* structureOutputSize */
                                                  &msrin,        /* inputStructure */
                                                  &msrout);       /* ouputStructure */
#else /* MAC_OS_X_VERSION_MIN_REQUIRED > 1040 */
        ret = IOConnectCallStructMethod(connect,
                                        AnVMSRActionMethodWRPMC,
                                        &msrin,
                                        sizeof(msrin),
                                        &msrout,
                                        &outsize);
#endif /* MAC_OS_X_VERSION_MIN_REQUIRED */

        if (ret != KERN_SUCCESS)
        {
            fprintf(stderr, "Can't connect to structure method to send commands.\n");

            retval = -1;
        } else {
            printf("WRPMC 0x%x with value 0x%llx.\n", (unsigned int)msrin.msr, (unsigned long long)msrin.param);
        }
    } else if (!strncmp(parameter, "writetsc", 8)) {
        if (argc < 3)
        {
            usage(argv[0]);

            retval = 1;

            return retval;
        }

        regvalue = (char *)argv[3];

        msrin.msr = 0;
        msrin.action = AnVMSRActionMethodWRTSC;
        msrin.param = (UInt64)hex2int(regvalue);

#if MAC_OS_X_VERSION_MIN_REQUIRED <= 1040
        ret = IOConnectMethodStructureIStructureO(connect,
                                                  AnVMSRActionMethodWRTSC,
                                                  sizeof(msrin),            /* structureInputSize */
                                                  &outsize,    /* structureOutputSize */
                                                  &msrin,        /* inputStructure */
                                                  &msrout);       /* ouputStructure */
#else /* MAC_OS_X_VERSION_MIN_REQUIRED > 1040 */
        ret = IOConnectCallStructMethod(connect,
                                        AnVMSRActionMethodWRTSC,
                                        &msrin,
                                        sizeof(msrin),
                                        &msrout,
                                        &outsize);
#endif /* MAC_OS_X_VERSION_MIN_REQUIRED */

        if (ret != KERN_SUCCESS)
        {
            fprintf(stderr, "Can't connect to structure method to send commands.\n");

            retval = -1;
        } else {
            printf("WRTSC with value 0x%llx.\n", (unsigned long long)msrin.param);
        }
    } else {
        usage(argv[0]);

        retval = 1;

        return retval;
    }

    if (connect != ((io_connect_t)0))
    {
        ret = IOServiceClose(connect);

        if (ret != KERN_SUCCESS)
        {
            fprintf(stderr, "Close IOService failed.\n");

            retval = -1;
        }
    }
   
    if (service != ((io_service_t)0))
    {
        ret = IOObjectRelease(service);

        if (ret != KERN_SUCCESS)
        {
            fprintf(stderr, "Release IOObject failed.\n");

            retval = -1;
        }
    }

    return retval;
}
